@@ -4,10 +4,19 @@
  *
  * Function: This program reads the contents of a flat file
  *           and loads it into a Mini SQL table.
+ *           The program does not create the MiniSQL table,
+ *           it must have been created beforehand.  msql-import
+ *           simply sends INSERTs to the database server.
  *
- * Date:     November 1995
+ *           When importing a flat fiel that was created in
+ *           the DOS world, make sure you convert the file
+ *           to unix by replacing CR/LF with LF only.
  *
- * Copyright (C) 1995 Pascal Forget.  All Rights Reserved
+ * Date:     March 1996
+ *
+ * Author:   Pascal Forget <pascal@wsc.com>
+ *
+ * Copyright (C) 1995-1996 Pascal Forget.  All Rights Reserved
  *
  * PASCAL FORGET MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY
  * OF THIS SOFTWARE FOR ANY PURPOSE.  IT IS SUPPLIED "AS IS"
@@ -22,7 +31,7 @@
 #include "msql-import.h"
 #include <string.h>
 
-#define MSQL_IMPORT_VERSION "0.0.6"
+#define MSQL_IMPORT_VERSION "0.0.9"
 #define IMPORT_MAX_FIELDS 255
 #define MAX(a,b) (a>b)a:b
 
@@ -88,7 +97,8 @@ append_string_buffer(SafeString *str, const char *newBuffer)
 void
 copy_string_buffer(SafeString *str, const char *newBuffer)
 {
-    if (newBuffer != NULL) {
+    if (newBuffer != NULL)
+    {
 	set_string_capacity(str, strlen(newBuffer));
 	strcpy(str[0].buffer, newBuffer);
     }
@@ -164,16 +174,23 @@ set_string_capacity(SafeString *str, unsigned int cap)
 char *
 safe_c_string_copy(const char *s)
 {
-	char	*s1;
-
-	if (!s)
-		return(NULL);
-	s1 = (char *)malloc(strlen(s) + 1);
-	if (!s1)
-		return(NULL);
-	
-	strcpy(s1,s);
-	return(s1);
+    char *s1;
+    
+    if (s == NULL)
+    {
+	return(NULL);
+    }
+    
+    s1 = (char *)malloc(strlen(s) + 1);
+    
+    if (s1 == NULL)
+    {
+	return(NULL);
+    }
+    
+    (void)strcpy(s1,s);
+    
+    return(s1);
 }
 
 /**********************************************************************/
@@ -208,10 +225,15 @@ str_parse( char **str, char *delimiters )
     char *head, *tail, *rv;
 
     if ( *str == NULL || **str == '\0' )
+    {
         rv = NULL;
+    }
     else if ( delimiters == NULL || *delimiters == '\0' )
+    {
         rv = NULL;
-    else {
+    }
+    else
+    {
         rv = head = *str;
         if (( tail = strpbrk( head, delimiters ) ) == NULL)
             *str = head + strlen(head);
@@ -277,9 +299,10 @@ append_string(char *s, const char *append)
 }
 
 /*****************************************************************
- * Put a backslash in front of single quote and parentheses      *
- * characters in the string.  This is needed for inserting       *
- * strings containing single quotes in a relational database.    *
+ * Put a backslash in front of single quote, parentheses,        *
+ * backslashes, and other special characters in the string. This *
+ * is needed for inserting strings containing these characters   *
+ * in a relational database.                                     *
  *****************************************************************/
 
 #define BACKSLASH_CHAR 92
@@ -288,12 +311,16 @@ char *
 backslashify_special_chars(const char *s)
 {
     char *str, *ptr;
-    int i, index=-2, len;
+    int i, j, index, len;
+
+    const char special_chars[] ={ '\'', '\"', 0}; 
+    int current = 0;
     
     if ((!s) ||
 	(strlen(s) == 0) ||
 	((strchr(s, '\'') == NULL) &&
 	 (strchr(s, ')') == NULL) &&
+	 (strchr(s, '\\') == NULL) &&
 	 (strchr(s, '(') == NULL)))
     {
 	return safe_c_string_copy(s);
@@ -301,51 +328,66 @@ backslashify_special_chars(const char *s)
 
     str = safe_c_string_copy(s);
     
-    while ((ptr = strchr(str+index+2, '\'')) != NULL)
+    /*
+     * Process the backslash character first.  We can't use strchr
+     * on strings looking for a backslash
+     */
+
+    i = 0;
+    
+    while (str[i] != '\0')
     {
-	len = strlen(str)+1;
-	str = (char *)realloc(str, len+1);
-	
-	index = ptr-str;
-
-	for (i=len+1; i>=index; i--)
+	if (str[i] == BACKSLASH_CHAR)
 	{
-	    str[i+1] = str[i];
+	    len = strlen(str) + 1;
+	    str = (char *)realloc(str, len+2);
+
+	    /* move the string to the right */
+	    
+	    j = len-1;
+
+	    while (j>i)
+	    {
+		str[j+2] = str[j--];
+	    }
+	    str[len+1] = '\0';
+
+	    /* insert 2 backslashes in the gap */
+
+	    str[i] = BACKSLASH_CHAR;
+	    str[i+1] = BACKSLASH_CHAR;
+	    
+	    i+=3; /* this is not a mistake */
 	}
-
-	str[index] = BACKSLASH_CHAR;
-    }
-
-    while ((ptr = strchr(str+index+2, '(')) != NULL)
-    {
-	len = strlen(str)+1;
-	str = (char *)realloc(str, len+1);
 	
-	index = ptr-str;
-
-	for (i=len+1; i>=index; i--)
+	if (str[i] != '\0')
 	{
-	    str[i+1] = str[i];
+	    i++;
 	}
-
-	str[index] = BACKSLASH_CHAR;
-    }
-
-    while ((ptr = strchr(str+index+2, ')')) != NULL)
-    {
-	len = strlen(str)+1;
-	str = (char *)realloc(str, len+1);
-	
-	index = ptr-str;
-
-	for (i=len+1; i>=index; i--)
-	{
-	    str[i+1] = str[i];
-	}
-
-	str[index] = BACKSLASH_CHAR;
     }
     
+    /* Do the other special characters */
+    
+    while(special_chars[current])
+    {
+	index = -2;
+
+	while ((ptr = strchr(str+index+2, special_chars[current])) != NULL)
+	{
+	    len = strlen(str)+1;
+	    index = ptr-str;
+	    str = (char *)realloc(str, len+1);
+	    
+	    for (i=len+1; i>=index; i--)
+	    {
+		str[i+1] = str[i];
+	    }
+	    
+	    str[index] = BACKSLASH_CHAR;
+	}
+	current++;
+    }
+
     return str;
 }
 
@@ -375,7 +417,9 @@ datatypes(const char *table, const char *fields)
     char **ptr;
     int done = 0;
     int found = 0;
-
+    const char *buf[] = {"unknown", "int", "char", "real", "ident", "null"};
+    const char **str = buf;
+    
     fieldsCopy = safe_c_string_copy(fields);
     ptr = &fieldsCopy;
 
@@ -394,7 +438,10 @@ datatypes(const char *table, const char *fields)
 		{
 		    if (!strcmp(field->name, token))
 		    {
-			types[fieldCount++] = field->type;
+			printf(" Field: %s ", field->name);
+			printf(" Type: %s\n", *(str+(int)field->type));
+			
+			types[fieldCount++ +1] = field->type;
 			found = 1;
 		    }
 		}
@@ -416,7 +463,10 @@ datatypes(const char *table, const char *fields)
 
 	while ((field = msqlFetchField(result)) != NULL)
 	{
-	    types[fieldCount++] = field->type;
+	    printf(" Field: %s\t", field->name);
+	    printf(" Type: %s\n", *(str+(int)field->type));
+
+	    types[fieldCount++ + 1] = field->type;
 	}
     }
     
@@ -428,11 +478,15 @@ datatypes(const char *table, const char *fields)
 	free(fieldsCopy);
     }
 #endif
+
+    /* Put the number of fields in the first position in the array */
     
+    types[0] = fieldCount;
+    
+    printf("\n");
     return types;
 }
 
-
 /*****************************************************************
  * _get_record reads one row from the data file and returns it   *
  *****************************************************************/
@@ -505,7 +559,8 @@ import_file(char *table,
 			      record[0].buffer,
 			      fields,
 			      data_types,
-			      recordCount++);
+			      recordCount++,
+			      fieldDel);
     }
 
     free(fieldDel);
@@ -555,20 +610,21 @@ import_file(char *table,
 
 int
 insert_record(const char *table,
-	      const char *record,
+	      char *record,
 	      const char *fields,
 	      int *types,
-	      int recordCount)
+	      int recordCount,
+	      char *fieldDel)
 {
     char **ptr = &record;
     char *token;
     char *value;
-    char c = '\t';
     char delimiter[2];
     int done = 0;
     int fieldNumber = 1;
+    int fieldsCount = types[0]; /* The number of values to be inserted */
     
-    sprintf(delimiter, "%c", c);
+    sprintf(delimiter, "%s", fieldDel);
     query[0].buffer[0] = '\0';
 
     copy_string_buffer(query, "INSERT INTO ");
@@ -587,7 +643,7 @@ insert_record(const char *table,
     
     if((record != NULL) && (token = str_parse(ptr,delimiter)))
     {
-	switch(types[0])
+	switch(types[1])
 	{
 	  case CHAR_TYPE:
 	    value = backslashify_special_chars(token);
@@ -620,7 +676,7 @@ insert_record(const char *table,
     {
 	if ((token = str_parse(ptr, delimiter)) != NULL)
 	{
-	    switch(types[fieldNumber++])
+	    switch(types[fieldNumber++ + 1])
 	    {
 	      case CHAR_TYPE:        
 		value = backslashify_special_chars(token);
@@ -649,6 +705,21 @@ insert_record(const char *table,
 	}
     }
 
+    /*
+     * Add NULL values if the number of values read in the record
+     * is lower than the number of field specified in the command
+     * line or if the number of values in the record is lower than
+     * the number of fields in the table if the field names weren't
+     * specified on the command line
+     */
+    
+    while (fieldNumber < fieldsCount)
+    {
+	append_string_buffer(query, "NULL,");
+	fieldNumber++;
+    }
+
+    
     /* Remove the last comma and the record delimiter ('\n') */
     query[0].buffer[strlen(query[0].buffer)-1] = '\0';
     
@@ -728,7 +799,7 @@ format_delimiter(const char *str)
 	return (char *)NULL;
     }
     
-    if (!strcmp(str, "\\n"))
+    if (strcmp(str, "\\n") == 0)
     {
 	result = malloc(2);
 	result[0] = nl;
@@ -736,7 +807,7 @@ format_delimiter(const char *str)
     }
     else
     {
-	if (!strcmp(str, "\\t"))
+	if (strcmp(str, "\\t") == 0)
 	{
 	    result = malloc(2);
 	    result[0] = tab;
@@ -756,24 +827,28 @@ printHelp(void)
 {
     const char text[] =
     {
-	"  msql-import loads the contents of a an ASCII delimited flat file "
-	"into a\n  MiniSQL table.  It automatically performs the type "
-	"conversions, and\n  validates the data.\n\n  msql-import is invoked "
-	"as follows:\n\n    msql-import host base tbl field_term "
-	"rec_term datafile [field [,field...]]\n\n"
+	"  msql-import loads the contents of an ASCII delimited flat file "
+	"into an\n  existing MiniSQL table.  It automatically performs the "
+	"type conversions, and\n  validates the data.\n\n  msql-import is "
+	"invoked as follows:\n\n    msql-import -h host -d database -t table "
+	"-c column_delimiter \\\n"
+	"    -r record_delimiter -i input_datafile "
+	"-f \"[field [,field...]]\"\n\n"
 	"               host: hostname of the msql server\n"
-	"               base: the database name\n"
-	"                tbl: the table in which to load the data\n"
-	"         field_term: the character used to delimit fields within a"
+	"           database: the Mini SQL database name\n"
+	"              table: the table in which to load the data\n"
+	"   column_delimiter: the character used to delimit fields within a"
 	" record\n"
-	"           rec_term: the character used to delimit records\n"
-	"           datafile: conatins the data to be imported\n"
+	"   record_delimiter: the character used to delimit records\n"
+	"     input_datafile: contains the data to be imported\n"
 	"             fields: import the data in those only (optional)\n\n"
-	"  Example:\n\n    msql-import zeus db table \\t \\n /tmp/file "
-	"client_id,name,address\n\n"
+	"  Example:\n\n    msql-import -h zeus -d db -t table -c \\t "
+	"-r \\n -i /tmp/file \\\n "
+	"    -f \"client_id,name,address\"\n\n"
 	"  If the fields are not specified, then all fields in the table "
 	"will be filled\n  with the data contained in the flat file, in "
 	"order of appearance.\n\n"
+	"  msql-import was written by Pascal Forget <pascal@wsc.com>.\n\n"
     };
 	
     fprintf(stdout, "\n  msql-import %s help:\n\n", MSQL_IMPORT_VERSION);
@@ -783,10 +858,22 @@ printHelp(void)
 void
 printUsage(void)
 {
-    fprintf(stderr, "Usage:\n"
-	    "       msql-import host base table field_terminator "
-	    "record_terminator datafile [field [, field...]]\n\n"
-	    "Type 'msql-import --help' for more information\n\n");
+    fprintf (stderr,
+	     "\nUsage: msql-import [-h host] [-c col_delimiter] "
+	     "[-r rec_delimiter]\n"
+	     "       -d database -t table -i input_file -f [field [,"
+	     " field...]]] \n\n"
+	     "       where host is host address or the hostname,"
+	     " default = localhost\n"
+	     "       col_delimiter is column delimiter, default = tab \n"
+	     "       rec_delimiter is record delimiter, default = newline\n"
+	     "       database is the mSQL database (mandatory)\n"
+	     "       table is the mSQL table (mandatory)\n"
+	     "       input_file is the input ascii text file (mandatory)\n"
+	     "       fields are optional input fields, must be in double"
+	     " quotes \n\n"
+	     "       Type 'msql-import --help' for more information\n\n");
+    exit (0);
 }
 
 void
@@ -796,7 +883,7 @@ printVersion(void)
 }
 
 /*****************************************************************
- * _main() connects to the server, then imports the data         *
+ * _main() connects to the msql server, then imports the data    *
  *****************************************************************/
 
 void
@@ -804,62 +891,147 @@ main(int argc, char **argv)
 {
     const char *buf[] ={"--help","--info","help","info","-help","-info",0};
     const char **ptr = buf;
-    const char *fields = NULL;
+    int len;
+    char *hostname = "localhost";
+    char *rec_delimiter = "\n";
+    char *col_delimiter = "\t";
+    char *database = NULL;
+    char *table = NULL;
+    char *input_file = NULL;
+    char *fields = NULL;
+    int i = argc - 1;
 
-    sock = -1;
-    
-    /* Validate calling arguments */
-
-    if (argc == 2)
-    {
-	if (!strcmp(argv[1], "--version"))
+    if (argc == 2) {
+	if ((strcmp(argv[1],"--version")==0) || (strcmp(argv[1],"--ver")==0))
 	{
 	    printVersion();
-	    return;
+	    exit(0);
 	}
 	else
 	{   
-	    while(ptr) {
-		if (!strcmp((char *)*ptr, argv[1]))
+	    while(*ptr)
+	    {
+		if (strcmp((char *)*ptr, argv[1]) == 0)
 		{
 		    printHelp();
-		    return;
+		    exit(0);
 		}
 		ptr++;
 	    }
+	    fprintf(stderr, "msql-import error: unknown option %s\n", argv[1]);
+	    printUsage();
 	}
     }
 
-    if (argc == 8)
+   /* Parse calling arguments */
+
+    while(i > 1)
     {
-	fields = safe_c_string_copy(argv[7]);
+	if (strcmp(argv[i-1], "-h") == 0)
+	{
+	    hostname = argv[i];
+	}
+	else if (strcmp(argv[i-1], "-c") == 0)
+	{
+	    col_delimiter = safe_c_string_copy(argv[i]);
+	}
+	else if (strcmp(argv[i-1], "-r") == 0)
+	{
+	    rec_delimiter = safe_c_string_copy(argv[i]);
+	}
+	else if (strcmp(argv[i-1], "-d") == 0)
+	{
+	    database = argv[i];
+	}
+	else if (strcmp(argv[i-1], "-t") == 0)
+	{
+	    table = safe_c_string_copy(argv[i]);
+	}
+	else if (strcmp(argv[i-1], "-i") == 0)
+	{
+	    input_file = safe_c_string_copy(argv[i]);
+	}
+	else if (strcmp(argv[i-1], "-f") == 0)
+	{
+	    fields = safe_c_string_copy(argv[i]);
+	}
+	i-=2;
+    }
+
+    /* check mandatory arguments. If one is missing, print usage then exit. */
+    
+    if ((input_file == NULL) || (table == NULL) || (database == NULL))
+    {
+	printUsage();
+    }
+
+    /* Print the arguments that were interpreted */
+    
+    printf ("\n hostname         =  %s\n", hostname);
+
+    if (strcmp(col_delimiter, "\t") !=0)
+    {
+	printf (" column delimiter =  %s\n", col_delimiter);
     }
     else
     {
-	if (argc != 7)
-	{
-	    printUsage();
-	    return;
-	}
+	printf (" column delimiter =  TAB\n");
     }
 
+    if (strcmp(rec_delimiter, "\n") !=0) {
+	printf (" record delimiter =  %s\n", rec_delimiter);
+    }
+    else
+    {
+	printf(" record delimiter =  NEWLINE\n");
+    }
+    
+    printf (" database         =  %s\n", database);
+    printf (" table            =  %s\n", table);
+    printf (" input_file       =  %s\n", input_file);
+
+    if (fields != NULL)
+    {
+	printf (" fields     =  %s\n", fields);
+    }
+
+    printf("\n");
+
     /* Connect to the msql server */
 
-    if (strcmp(argv[1], "localhost") == 0) {
+    if (strcmp(hostname, "localhost") == 0)
+    {
 	sock = msqlConnect(NULL);
-    } else {
-	sock = msqlConnect(argv[1]);
+    }
+    else
+    {
+	sock = msqlConnect(hostname);
     }
 
-    /* Set the current database */
-    
-    if ((sock == -1) || (msqlSelectDB(sock, argv[2]) == -1)) {
+    /* Verify that the connection has been successfully established */
+
+    if (sock == -1)
+    {
+	fprintf(stderr, "msql-import: error connecting to host. "
+		"Exiting.\n");
 	abort_import(1);
-    }    
-	
-    import_file(safe_c_string_copy(argv[3]),
-		safe_c_string_copy(argv[6]),
-		format_delimiter(safe_c_string_copy(argv[4])),
-		format_delimiter(safe_c_string_copy(argv[5])),
-		fields);    
+    }
+    
+    /* Set the current database */
+
+    if (msqlSelectDB(sock, database) == -1)
+    {
+	fprintf(stderr, "msql-import: error opening database. "
+		"Exiting.\n");
+	abort_import(1);
+    }
+    
+    /* ssk test if table exists */
+    len = row_length( table );
+    
+    import_file(table,
+		input_file,
+		format_delimiter(col_delimiter),
+		format_delimiter(rec_delimiter),
+		fields);
 }
